import os
import os.path
import hashlib
from urllib.parse import urlparse
from git import Repo
from pathlib import Path

GIT_CHACHE_BASE_PATH = '~/.cache/imka.d4rk.io'

class FileProvider:
    basePath: str

    def __init__(self, basePath):
        self.basePath = basePath

    def isdir(self, path):
        return os.path.isdir(self.get_real_path(path))

    def exists(self, path):
        return os.path.exists(self.get_real_path(path))

    def walk(self, path):
        dir_files = []

        for root, dirs, files in os.walk(self.get_real_path(path), topdown=False):
            for name in files:
                real_path = os.path.join(root, name)
                frame_path = os.path.relpath(real_path, self.basePath)
                dir_files.append(frame_path)

        return dir_files

    def open(self, path, mode='r'):
        return open(self.get_real_path(path), mode)

    def get_real_path(self, path):
        return os.path.join(self.basePath, path)

class GitFileProvider(FileProvider):
    def __init__(self, frame_url, version):
        self.version = version
        if not self.version:
            self.version = 'main'

        url = urlparse(frame_url)
        scheme = {'git+https': 'https', 'git+ssh': 'ssh', 'git+file': 'file'}.get(url.scheme)
        self.repo_url = url._replace(scheme=scheme, fragment='').geturl()

        url_hash = hashlib.sha256(self.repo_url.encode()).hexdigest()
        self.repo_path = os.path.expanduser(os.path.join(GIT_CHACHE_BASE_PATH, url_hash[:2],url_hash))

        if not os.path.exists(self.repo_path):
            self.repo = Repo.clone_from(self.repo_url, to_path=self.repo_path)
        else:
            self.repo = Repo(self.repo_path)

        self.repo.head.reference = self.repo.commit(self.version)
        self.repo.head.reset(index=True, working_tree=True)

        self.basePath = os.path.join(self.repo_path, url.fragment)
        if not os.path.exists(self.basePath):
            raise Exception('Frame subpath ({}) in repo dose not exist - {}'.format(url.fragment, self.basePath) )
        
        self.basePath


import yaml
from urllib.parse import urlparse

from .file_provider import FileProvider, GitFileProvider

COMPOSE_VERSION_POSTFIX = '-imka/v1'

class FrameController:
    templateController: object
    imkaConfigController: object
    stackController: object

    def __init__(self, templateController, imkaConfigController, stackController):
        self.templateController = templateController
        self.imkaConfigController = imkaConfigController
        self.stackController = stackController
        
    def load_frame_from_uri(self, frame_url, version):
        url = urlparse(frame_url)

        if url.scheme == '':
            fileProvider = FileProvider(frame_url)
        elif url.scheme in ['git+https', 'git+ssh', 'git+file']:
            fileProvider = GitFileProvider(frame_url, version)
        else:
            raise Exception("unknown url schema for frame")

        with fileProvider.open('frame.yml') as file:
            config = yaml.safe_load(file)

            compose_templates = config['compose_templates']
            name = config['name']

            valueFile = None
            if fileProvider.exists('values.yml'):
                valueFile = fileProvider.get_real_path('values.yml')

        return Frame(fileProvider, compose_templates, name, valueFile)

    def render_templates(self, frame, values):
        rendered = self.templateController.render_compose_templates(frame, values)

        for cfg in rendered:
            print('Rendering config {}'.format(cfg))

    def evaluate_compose_yml(self, frame, values):
        self.render_templates(frame, values)

        if not frame.compose_yml['version'].endswith(COMPOSE_VERSION_POSTFIX):
            raise Exception('compose version must end with {}'.format(COMPOSE_VERSION_POSTFIX))

        self.imkaConfigController.evaluate_compose_yml(frame, values)

        frame.compose_yml['version'] = frame.compose_yml['version'].replace(COMPOSE_VERSION_POSTFIX, '')

    def apply(self, frame, values):
        self.evaluate_compose_yml(frame, values)

        created = self.imkaConfigController.docker_create_configs(values)
        for cfg in created:
            print('Creating config {}'.format(cfg))

        self.stackController.apply(frame, values)

        deleted = self.imkaConfigController.docker_delete_old_config_versions()
        for cfg in deleted:
            print('Removing config {}'.format(cfg))

    def down(self, values):
        self.stackController.down(values)
        deleted = self.imkaConfigController.docker_delete_all_configs(values)

        for cfg in deleted:
            print('Removing config {}'.format(cfg))

class Frame:
    fileProvider: None
    compose_yml: dict
    compose_templates: list
    value_file: str
    name: str

    def __init__(self, fileProvider, compose_templates, name, value_file):
        self.fileProvider = fileProvider
        self.compose_templates = compose_templates
        self.name = name
        self.value_file = value_file

    def get_compose_templates(self):
        paths = []

        for path in self.compose_templates:
            paths.append(path)

        return paths


import click
import yaml

from .imka import ImkaController

@click.group()
@click.pass_context
def main(ctx):
    ctx.obj = ImkaController()

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesting depth')
@click.option('--version', type=str, help='specify a frame version, only works for git: tag, branch or commit')
@click.pass_context
def values(ctx, frame, deployment, values, render_values_depth, version):
    dump = yaml.dump(ctx.obj.load_values(frame, deployment, values, render_values_depth, version))
    print('---')
    print(dump)

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesting depth')
@click.option('--version', type=str, help='specify a frame version, only works for git: tag, branch or commit')
@click.pass_context
def template(ctx, frame, deployment, values, render_values_depth, version):
    ctx.obj.render_templates(frame, deployment, values, render_values_depth, version)
    print('---')
    print(yaml.dump(ctx.obj.chart.compose_yml))

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesting depth')
@click.option('--version', type=str, help='specify a frame version, only works for git: tag, branch or commit')
@click.pass_context
def apply(ctx, frame, deployment, values, render_values_depth, version):
    ctx.obj.apply(frame, deployment, values, render_values_depth, version)

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesting depth')
@click.option('--version', type=str, help='specify a frame version, only works for git: tag, branch or commit')
@click.pass_context
def down(ctx, frame, deployment, values, render_values_depth, version):
    ctx.obj.down(frame, deployment, values, render_values_depth, version)


import jinja2
import yaml

class TemplateController:
    def render_template(self, content, values):
        template = jinja2.Template(content)

        return template.render(values)

    def render_compose_templates(self, frame, values):
        rendered = []
        compose = {}

        for path in frame.get_compose_templates():
            with frame.fileProvider.open(path) as file:
                content = file.read()
            
            compose = self._merge_yaml(compose, yaml.safe_load(self.render_template(content, values)))

            rendered.append(path)

        frame.compose_yml = compose

        return rendered

    def _merge_yaml(self, destination, source):
        for key, value in source.items():
            if isinstance(value, dict) and value:
                node = destination.get(key, {})
                destination[key] = self._merge_yaml(node, value)
            else:
                destination[key] = value

        return destination


from dataclasses import dataclass
import hashlib
import os

CONFIG_TYPE = 'imka'
CONFIG_ID_LABEL = 'd4rk.io/imka/config/id'
CONFIG_BELONGS_TO_LABEL = 'd4rk.io/imka/config/belongs_to'
DEPLOYMENT_LABEL = 'd4rk.io/imka/deployment'

class ImkaConfigController:
    configs: list
    templateController = None
    dockerClient = None

    def __init__(self, dockerClient, templateController):
        self.configs = []
        self.dockerClient = dockerClient
        self.templateController = templateController

    def _add_config(self, frame, values, sourcePath, targetPath, imka_templating, mount_uid, mount_gid, mount_mode):
        if not frame.fileProvider.isdir(sourcePath):
            return [
                self._create_config(frame, values, sourcePath, targetPath, imka_templating, mount_uid, mount_gid, mount_mode, None)
            ]

        configId = self._generate_config_id(values, sourcePath)

        configs = []
        for path in frame.fileProvider.walk(sourcePath):
            pathRelativeToSource = os.path.relpath(path, sourcePath)
            mountPath = os.path.join(targetPath, pathRelativeToSource)

            config = self._create_config(frame, values, path, mountPath, imka_templating, mount_uid, mount_gid, mount_mode, configId)
            configs.append(config)

        return configs

    def _create_config(self, frame, values, sourcePath, targetPath, imka_templating, mount_uid, mount_gid, mount_mode, belongs_to):
        data = self._read_data(frame, values, sourcePath, imka_templating)

        config = ImkaConfig(
            self._generate_config_id(values, sourcePath),
            self._generate_version_id(data, values),
            targetPath,
            mount_uid,
            mount_gid,
            mount_mode,
            belongs_to,
            data
        )

        self.configs.append(config)

        return config

    def _read_data(self, frame, values, sourcePath, imka_templating):
        if imka_templating != True:
            with frame.fileProvider.open(sourcePath, 'rb') as file:
                return file.read()

        with frame.fileProvider.open(sourcePath) as file:
            content = file.read()

        rendered = self.templateController.render_template(content, values)
        return rendered.encode()

    def _generate_config_id(self, values, sourcePath):
        localId = hashlib.sha256(sourcePath.encode()).hexdigest()[:32]

        return '{}-{}'.format(values['deployment_fullname'], localId)

    def _generate_version_id(self, data, values):
        return hashlib.sha256(values['deployment_fullname'].encode() + data).hexdigest()

    # before apply
    def docker_create_configs(self, values):
        modified = []

        for config in self.configs:
            configs = self.dockerClient.configs.list(filters={"name": config.versionId})

            if len(configs) != 0:
                continue # config with this version already exists

            labels = {
                DEPLOYMENT_LABEL: values['deployment_fullname'], # use to delete all configs on down
                CONFIG_ID_LABEL: config.configId,
            }

            if config.belongs_to:
                labels[CONFIG_BELONGS_TO_LABEL] = config.belongs_to

            self.dockerClient.configs.create(name=config.versionId, data=config.data, labels=labels)

            modified.append(config.versionId)

        return modified

    # after apply
    def docker_delete_old_config_versions(self):
        modified = []

        for config in self.configs:
            configs = self.dockerClient.configs.list(filters={"label": '{}={}'.format(CONFIG_ID_LABEL, config.configId)})

            for live_config in configs:
                if live_config.name != config.versionId:
                    live_config.remove()
                    modified.append(live_config.name)

        return modified

    # down
    def docker_delete_all_configs(self, values):
        configs = self.dockerClient.configs.list(filters={"label": 'd4rk.io/imka/deployment={}'.format(values['deployment_fullname'])})

        removed = []
        for config in configs:
            config.remove()
            removed.append(config.name)

        return removed

    def evaluate_compose_yml(self, frame, values):
        if 'configs' not in frame.compose_yml:
            frame.compose_yml['configs'] = {}

        compose_configs = frame.compose_yml['configs']

        for serviceName, service in frame.compose_yml.get('services', {}).items():

            serviceConfigs = service.get('configs', [])

            for serviceConfigIndex in reversed(range(len(serviceConfigs))): # iterate list backwared, because items need to be inserted at the current index | is there a better way to do that in python?
                serviceConfig = serviceConfigs[serviceConfigIndex]
                if isinstance(serviceConfig, str) or serviceConfig.get('type', '') != CONFIG_TYPE:
                    continue

                sourcePath = serviceConfig.get('source', False)
                targetPath = serviceConfig.get('target', False)
                imka_templating = serviceConfig.get('template', {}).get('enabled', False)
                mount_uid = serviceConfig.get('uid', False)
                mount_gid = serviceConfig.get('gid', False)
                mount_mode = serviceConfig.get('mode', False)

                if not sourcePath:
                    raise Exception('services.{}.configs[{}].source must be set'.format(serviceName, serviceConfigIndex))

                if not targetPath:
                    raise Exception('services.{}.configs[{}].target must be set'.format(serviceName, serviceConfigIndex))

                configs = self._add_config(frame, values, sourcePath, targetPath, imka_templating, mount_uid, mount_gid, mount_mode)

                serviceConfigs.pop(serviceConfigIndex)

                for config in configs: # if configs need a specific order, the list needs to be reversed hear
                    serviceConfigs.insert(serviceConfigIndex, config.get_service_config_object())
                    compose_configs[config.configId] = config.get_config_object()


@dataclass
class ImkaConfig:
    """
    Represents a singe config object
    """
    configId: str # configId is used to identify a config independent of its version
    versionId: str # hash of templated file content
    targetPath: str
    mount_uid: str
    mount_gid: str
    mount_mode: int
    belongs_to: str # configId of parrent config if the source is a dir
    data: bytes

    def get_config_object(self):
        return {
            "external": True,
            "name": self.versionId
        }

    def get_service_config_object(self):
        return {
            "target": self.targetPath,
            "source": self.configId
            #todo mount opts
        }


import yaml
import sys

from subprocess import Popen, PIPE, STDOUT

class StackController:
    def apply(self, frame, values):
        name = values['deployment_fullname']

        dockerStack = frame.compose_yml
        dockerStackYaml = yaml.dump(dockerStack)
        dockerStackYamlEncoded = dockerStackYaml.encode()

        p = Popen(['docker', 'stack', 'deploy', '-c', '-', name], stdin=PIPE, stderr=sys.stderr.buffer, stdout=sys.stdout.buffer)
        p.communicate(input=dockerStackYamlEncoded)[0]

    def down(self, values):
        name = values['deployment_fullname']

        p = Popen(['docker', 'stack', 'rm', name], stderr=sys.stderr.buffer, stdout=sys.stdout.buffer)
        p.communicate()[0]


import docker

from .frames import FrameController
from .templates import TemplateController
from .imka_configs import ImkaConfigController
from .stack import StackController
from .values import ValueController

class ImkaController:
    frameController: object
    valueController: object

    def __init__(self,):
        self.frameController = FrameController(
            TemplateController(),
            ImkaConfigController(docker.from_env(), TemplateController()),
            StackController(),
        )

        self.valueController = ValueController()

    def _init_frame(self, frame, version):
        self.frame = self.frameController.load_frame_from_uri(frame, version)

    def load_values(self, frame, deployment, value_files, render_values_depth, version):
        self._init_frame(frame, version)

        self.values = {
            'deployment': deployment,
            'deployment_fullname': '{}-{}'.format(self.frame.name, deployment)
        }

        self.imka_opts = {
            'render_values_depth': render_values_depth
        }

        self.values = self.valueController.load_values(self.frame, self.values, value_files, self.imka_opts)

        return self.values

    def render_templates(self, frame, deployment, value_files, render_values_depth, version):
        self.load_values(frame, deployment, value_files, render_values_depth, version)

        return self.frameController.evaluate_compose_yml(self.frame, self.values)

    def apply(self, frame, deployment, value_files, render_values_depth, version):
        self.load_values(frame, deployment, value_files, render_values_depth, version)
        return self.frameController.apply(self.frame, self.values)
        

    def down(self, frame, deployment, value_files, render_values_depth, version):
        self.load_values(frame, deployment, value_files, render_values_depth, version)
        self.frameController.down(self.values)


import jinja2
import yaml

class ValueController:
    def load_values(self, frame, values, value_files, imka_opts):
        if frame.value_file:
            value_files = [frame.value_file] + list(value_files)

        for path in value_files:
            with open(path) as file:
                current = yaml.safe_load(file)
            values = self._merge_yaml(values, current)

        depth = imka_opts.get('render_values_depth', 32)
        for i in range(depth):
            values = self._render_values(values, values, i == depth-1)

        return values

    def _render_values(self, values, node, last):
        new = {}

        for key, value in node.items():
            if isinstance(value, dict) and value:
                new[key] = self._render_values(values, value, last)
            elif isinstance(value, str):
                template = jinja2.Template(value)
                new[key] = template.render(values)

                if last and new[key].find('{{') >= 0:
                    raise Exception('j2 templates could not be fully resolved!')
            else:
                new[key] = value

        return new

    def _merge_yaml(self, destination, source):
        for key, value in source.items():
            if isinstance(value, dict) and value:
                node = destination.get(key, {})
                destination[key] = self._merge_yaml(node, value)
            else:
                destination[key] = value

        return destination